/*
 * This souce code is public domain. Enjoy!
 * http://hildstrom.com/projects/buellstator
 *
 * Last update: 201205021920
 *
 * 8-channel data logger for Arduino Uno
 *
 * From sensors:
 * stator thermocouple, regulator thermistor, air thermistor, battery voltage
 *
 * From Buell ECM:
 * rpm, speed, engine temperature, fan duty cycle
 *
 * PINS
 * SPI data output -> D2 -> MAX6675-1 DO
 * SPI chip select -> D3 -> MAX6675-1 CS
 * SPI clock -> D4 -> MAX6675-1 CLK
 * +5V -> 10K thermistor -> A0 -> 10K resistor -> GND
 * +5V -> 10K thermistor -> A1 -> 10K resistor -> GND
 * +15V -> 20K resistor -> A2 -> 10K resistor -> GND
 * software serial RX -> D8 -> Buell ECM TX
 * software serial TX -> D9 -> Buell ECM RX
 *
 * EEPROM layout
 * [0-1015] interleaved 8-bit scan data st,rt,at,bv,rpm,spd,ets,fdc,st,rt,at,bv,rpm,spd,ets,fdc...
 * [1022] init byte
 * [1023] next scan number
 * 1016 bytes / 8 sensors = 127 scans for 0-126
 *
 * The logger will save scans and revolve to overwrite old scans.
 * Scan 127 occupies addresses 126*8=1008 to 1015 in the EEPROM.
 * The very next scan should be 0. The variables should make it easy to
 * adapt to other EEPROM sizes or number of sensors.
 * The log_delay >= display_delay.
 */

//includes
#include <EEPROM.h> //for EEPROM I/O
#include "max6675.h" //for MAX6675 thermocouple IC
#include <math.h>  //for thermistor temperature
#include <SoftwareSerial.h> //for serial ecm communication

//global variables
//user-defined
#define use_eeprom 1
#define num_sensors 8
const char* sensor_names[num_sensors]={
  "stator_thermocouple",
  "regulator_thermistor",
  "air_thermistor",
  "battery_voltage",
  "RPM",
  "VSS",
  "CLT",
  "Fan_Duty"
};
#define first_address 0
#define scan_address 1023
#define init_value 7
#define display_delay 5000
#define log_delay 60000
#define stator_tc_do_pin 2
#define stator_tc_cs_pin 3
#define stator_tc_clk_pin 4
#define regulator_temp_pin 0
#define air_temp_pin 1
#define batt_volt_pin 2
#define ecm_rx 8
#define ecm_tx 9
//internal
const unsigned short init_address = scan_address-1;
const unsigned short num_scans = (scan_address-1-first_address)/num_sensors;
unsigned short scan = 0;
unsigned short sensor = 0;
unsigned short address = 0;
unsigned long current_milliseconds = 0;
unsigned long previous_display_milliseconds = 0;
unsigned long previous_log_milliseconds = 0;
MAX6675 stator_tc(stator_tc_clk_pin,stator_tc_cs_pin,stator_tc_do_pin);
byte sensor_values[num_sensors]={
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0
};
/*
 * ECM request for run-time data
 * from http://ecmspy.com/eeprom_info.shtml
 * from http://arduinobuell.blogspot.com/2012/03/good-news.html
 */
#define ecm_request_size 9
const byte ecm_request[ecm_request_size]={
  0x01, //soh: start of header
  0x00, //emmitend: from PC
  0x42, //recipient: to ECM
  0x02, //data size: 2 bytes
  0xFF, //eoh: end of header
  0x02, //sot: start of text
  0x43, //data 1: command code 56 get version 43 get run data
  0x03, //data 2: eot: end of text
  0xFD}; //checksum: xor checksum bytes 1-8 E8 get version FD get run data
/*
 * ECM response for run-time data
 * from http://ecmspy.com/cgi-bin/runtime.cgi
 */
#define ecm_response_size 135
byte ecm_response[ecm_response_size];
byte ecm_bytes = 0;
byte ecm_checksum = 0;
SoftwareSerial ecm(ecm_rx,ecm_tx);

//function to convert a 10k thermistor adc value to temperature
//from http://arduino.cc/playground/ComponentLib/Thermistor2
double thermistor(int RawADC) {
 double Temp;
 Temp = log(((10240000/RawADC) - 10000));
 Temp = 1 / (0.001129148 + (0.000234125 + (0.0000000876741 * Temp * Temp ))* Temp );
 Temp = Temp - 273.15;            // Convert Kelvin to Celsius
 //Temp = (Temp * 9.0)/ 5.0 + 32.0; // Convert Celsius to Fahrenheit
 return Temp;
}

//function to convert double to byte
byte double_to_byte(double d){
  if(d>255){
    return 255;
  }
  else if(d<0){
    return 0;
  }
  else{
    return (byte)d;
  }
}

//setup function
void setup(){
  //setup serial ports
  Serial.begin(9600);
  ecm.begin(9600);
  
  //setup i/o here
  
  //only dump the eeprom if we are using the eeprom
  if(use_eeprom){
    //wait 5 seconds
    delay(5000);
    Serial.println("LOGDATASTART");
    
    //check if the eeprom has been initialized before
    if(EEPROM.read(init_address) != init_value){
      for(address=first_address; address<=scan_address; address++){
        EEPROM.write(address,0);
      }
      EEPROM.write(init_address,init_value);
    }
    
    //dump the eeprom to the computer over serial/usb
    Serial.print("scan");
    Serial.print("\t");
    for(sensor=0; sensor<num_sensors; sensor++){
      Serial.print(sensor_names[sensor]);
      if(sensor != num_sensors-1){
        Serial.print("\t");
      }
    }
    Serial.println();
    for(scan=0; scan<num_scans; scan++){
      address=scan*num_sensors;
      Serial.print(scan,DEC);
      Serial.print("\t");
      for(sensor=0; sensor<num_sensors; sensor++){
        Serial.print(EEPROM.read(address+sensor),DEC);
        if(sensor != num_sensors-1){
          Serial.print("\t");
        }
      }
      Serial.println();
    }
    Serial.println("LOGDATAEND");
    Serial.println();
    Serial.println();
    Serial.println();
    Serial.println();
    Serial.println();
    
    //read the next scan number stored in the EEPROM
    scan = EEPROM.read(scan_address);
    
    //wait 5 seconds
    delay(5000);
  }
  
  //setup start time
  current_milliseconds = millis();
  previous_display_milliseconds = current_milliseconds;
  previous_log_milliseconds = current_milliseconds;
  
  //serial/usb outupt header
  Serial.print("scan");
  Serial.print("\t");
  for(sensor=0; sensor<num_sensors; sensor++){
    Serial.print(sensor_names[sensor]);
    if(sensor != num_sensors-1){
      Serial.print("\t");
    }
  }
  Serial.println();
}

//loop function
void loop(){
  //get elapsed time
  current_milliseconds = millis();
  
  //read once per display delay
  if(current_milliseconds - previous_display_milliseconds > display_delay){
    previous_display_milliseconds = current_milliseconds;
    
    //perform sensor measurements
    sensor_values[0] = double_to_byte(stator_tc.readCelsius());//stator temperature
    sensor_values[1] = double_to_byte(thermistor(analogRead(regulator_temp_pin)));//regulator temperature
    sensor_values[2] = double_to_byte(thermistor(analogRead(air_temp_pin)));//air temperature
    sensor_values[3] = analogRead(batt_volt_pin)/4;//battery voltage adc 10-bit to 8-bit
    
    //get ecm run-time data via serial
    //clear the recieve buffer
    ecm.flush();
    //listen for data
    ecm.listen();
    //send run-time data request
    for(ecm_bytes=0; ecm_bytes<ecm_request_size; ecm_bytes++){
      ecm.write(ecm_request[ecm_bytes]);
    }
    //receive run-time data response
    ecm_bytes = 0;
    ecm_response[0] = 0;
    while(ecm_bytes < ecm_response_size){
      if(ecm.available()){
        ecm_response[ecm_bytes] = ecm.read();
	if(ecm_response[0] == 1){
          ecm_bytes++;
	}
      }
      else{
      	if(millis()-previous_display_milliseconds > display_delay){
	  break;
	}
      }
    }
    //check for correct number of bytes
    if(ecm_bytes == ecm_response_size){
      //calculate checksum
      ecm_checksum = 0;
      for(ecm_bytes=1; ecm_bytes<ecm_response_size-1; ecm_bytes++){
        ecm_checksum ^= ecm_response[ecm_bytes];
      }
      //compare checksums
      if(ecm_checksum == ecm_response[ecm_response_size-1]){
      	Serial.println("#success: ecm response checksums matched");
        //extract values of interest from the ecm response
        sensor_values[4] = ((unsigned long)((unsigned long)ecm_response[12] << 8 | (unsigned long)ecm_response[11]))*255/12000;//engine speed in rpm *255/12000
        sensor_values[5] = ecm_response[99];//speed in kph
        sensor_values[6] = ((unsigned long)((unsigned long)ecm_response[31] << 8 | (unsigned long)ecm_response[30]))/10-40;//engine temperature C
        sensor_values[7] = ecm_response[98];//fan duty cycle
      }
      else{
        Serial.println("#error: ecm response checksums did not match");
      }
    }
    else{
      Serial.println("#error: incorrect number of ecm response bytes");
    }
    
    //output results
    Serial.print(scan,DEC);
    Serial.print("\t");
    for(sensor=0; sensor < num_sensors; sensor++){
      Serial.print(sensor_values[sensor],DEC);
      if(sensor != num_sensors-1){
        Serial.print("\t");
      }
    }
    Serial.println();
  }
  
  //log once per log delay
  if(current_milliseconds - previous_log_milliseconds > log_delay && use_eeprom){
    previous_log_milliseconds = current_milliseconds;
    
    //save results
    address=scan*num_sensors;
    for(sensor=0; sensor < num_sensors; sensor++){
      EEPROM.write(address+sensor,sensor_values[sensor]);
    }
    
    //revolve back to beginning of eeprom address space when necessary
    if(scan >= num_scans-1){
      scan=0;
    }
    else{
      scan++;
    }
    EEPROM.write(scan_address,scan);
  }
  
}

